/*
 * sbc2440-adc.c
 
A/D Converter
ADCCON 0x58000000 ADC R/W ADC control
ADCTSC 0x58000004 ADC touch screen control
ADCDLY 0x58000008 ADC start or interval delay
ADCDAT0 0x5800000C R ADC conversion data
ADCDAT1 0x58000010 ADC conversion data
ADCUPDN 0x58000014 R/W Stylus up or down interrupt status

3. All registers except ADC registers, RTC registers and UART registers must be read/write in word unit (32-bit) in
little/big endian.
4. Make sure that the ADC registers, RTC registers and UART registers be read/write by the specified access unit
and the specified address. Moreover, one must carefully consider which endian mode is used.
5. W : 32-bit register, which must be accessed by LDR/STR or int type pointer (int *).
HW : 16-bit register, which must be accessed by LDRH/STRH or short int type pointer (short int *).
B : 8-bit register, which must be accessed by LDRB/STRB or char type pointer (char int *).

*/

#if !defined(__KERNEL__)
#define	__KERNEL__
#endif

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/delay.h>
#include <linux/wait.h>
#include <linux/poll.h>
#include <asm/delay.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/ioctl.h>
#include <asm/arch/regs-adc.h>
#include <asm/arch/regs-gpio.h>
#include <asm/hardware/clock.h>

typedef union {
	unsigned short WORD;
	struct {
		unsigned char H:8;
		unsigned char L:8;
	} BYTE;
} _ADDR;


#define ADC_DEVNAME	"sbc2440_adc"
#define ADC_MAJOR	191
#define ADC_IRQ	23

#define  ECFLAG     0x8000
	// ǥХɥ饤Ͽ00
#define  SEL_MUX    0x38
#define  STDBM      0x04
#define  READ_START 0x02
#define  ENB_START  0x01

MODULE_AUTHOR( "ohshima@popkids.co.jp" );
MODULE_DESCRIPTION( "QQ2440 ADC driver." );
MODULE_LICENSE( "GPL" );

static unsigned int adc_major = ADC_MAJOR;
static char *adc_devname = ADC_DEVNAME;
static unsigned char *msg = "sbc2440_adc.o";

static unsigned int adc_ch = 0x00;
static int f_adc = 0;
static _ADDR	ad_data;

static void __iomem *base_addr;

// 
static int f_access = 0;

// ɤ߽Фݥ
static int read_p = 0;


// ߽

void adc_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	volatile unsigned char *addr = base_addr+S3C2410_ADCDAT0;
	unsigned char d;

	// ɶɤ
	d = *addr;
	// ADF & ADIE & ADST ꥢ
	*addr &= 0x1F;
	addr = addr + (adc_ch * 2);
	
	// ǡɤߤ̡̤νɤ߼
	ad_data.BYTE.H = *addr;
	addr++;
	ad_data.BYTE.L = *addr;
	// 6ӥåȥե
	ad_data.WORD = ad_data.WORD >> 6;
	// λե饰å
	f_adc = 1;
}



static int adc_open(struct inode *inode, struct file *filp)
{
	// 1ץץǤʤ褦ˤ
	if( f_access ) {
		printk( "%s: ADC is already used.", msg );
		return -EFAULT;
	}
	
	adc_ch = MINOR(inode->i_rdev);
	if( adc_ch > 3 ) {
		printk( "%s: ch > 3 is not supported.", msg );
		return -EFAULT;
	}

base_addr=ioremap(S3C2410_PA_ADC,0x20);
if (base_addr == NULL) {
	printk(KERN_ERR "Failed to remap register block\n");
	return -ENOMEM;
}
	
	read_p = 0;
	f_access++;
//	MOD_INC_USE_COUNT;
	return 0;
}

// release

static int adc_release(struct inode *inode, struct file *filp)
{
	f_access--;
	
//	MOD_DEC_USE_COUNT;
	return 0;
}


static ssize_t
adc_read(struct file *filp, char *buf, size_t count, loff_t *f_pos)
{
	volatile unsigned *adccon;
//base_addr=ioremap(S3C2410_PA_ADC,0x20);
//if (base_addr == NULL) {
//	printk(KERN_ERR "Failed to remap register block\n");
//	return -ENOMEM;
//}
adccon = base_addr+S3C2410_ADCCON;

	if( read_p == 0 )
	{
		// ñ⡼ADC
		*adccon &= ~STDBM;
		*adccon |= READ_START;
		*adccon |= ENB_START;

printk( KERN_INFO "S3C2410_ADCCON=> %04x \n", *adccon );
	
		while(1){
			if ( *adccon & ECFLAG ) break;
		}
	}
printk( KERN_INFO "S3C2410_ADCCON=> %04x \n", *adccon );
/*
	if( read_p == 0 ) {
		// ñ⡼ADC
		f_adc = 0;
		*adccon |= adc_ch;
		// Ԥ
		while( ! f_adc )
			schedule();
	}
*/	
	if( count >= 2 ) {
		buf[0] = ad_data.BYTE.H;
		buf[1] = ad_data.BYTE.L;
		read_p = 2;
	}
	else if( count != 0 ) {
		if( read_p == 0 ) {
			buf[0] = ad_data.BYTE.H;
			read_p++;
		}
		else {
			buf[0] = ad_data.BYTE.L;
			read_p++;
		}
	}
	if( read_p > 1 ){
		read_p = 0;
	}

	return read_p;
}


struct file_operations adc_fops = {
	owner:   THIS_MODULE,
	read:    adc_read,
	write:   NULL,
	open:    adc_open,
	release: adc_release,
	ioctl:   NULL,
};


// init
static struct clk	*adc_clock;

static int __init
adc_init(void)
{
	int retval;
unsigned long data_con;
unsigned long data0;
		
	// ǥХɥ饤Ͽ
	retval = register_chrdev(adc_major, adc_devname , &adc_fops);
	if (retval < 0) {
    	printk(KERN_ERR "Cannnot register: %s(%d)\n", adc_devname, adc_major);
    	return retval;
	}
	else {
    	printk(KERN_INFO "%s(%d): registered.\n", adc_devname, adc_major );

adc_clock = clk_get(NULL, "adc");
if (!adc_clock) {
	printk(KERN_ERR "failed to get adc clock source\n");
	return -ENOENT;
}
clk_use(adc_clock);
clk_enable(adc_clock);

base_addr=ioremap(S3C2410_PA_ADC,0x20);
if (base_addr == NULL) {
	printk(KERN_ERR "Failed to remap register block\n");
	return -ENOMEM;
}
/*
writel( 0xBFC4, base_addr+S3C2410_ADCCON );
data0 = readl(base_addr+S3C2410_ADCDAT0);
*/

//writel( 0x3FC4, base_addr+S3C2410_ADCCON );

data_con = readl(base_addr+S3C2410_ADCCON);

    	printk( KERN_INFO "S3C2410_ADCCON=> %04x \n", data_con );
  	}
/*
	retval = request_irq( ADC_IRQ, adc_interrupt, SA_SAMPLE_RANDOM, adc_devname, NULL );
	if( retval < 0 ) {
		printk(KERN_ERR "%s(%d): Cannot get irq %d.\n", adc_devname, adc_major,ADC_IRQ );
		iounmap(base_addr);
		return retval;
	}
*/
	writel((unsigned int)0, base_addr+S3C2410_ADCTSC);

	f_access = 0;
	adc_ch = 0;
	read_p = 0;
	
  	return 0;
}

// exit

static void __exit
adc_cleanup(void)
{
	// ߲
	free_irq( ADC_IRQ, NULL );
  	printk( KERN_INFO "%s : removed.\n", msg );
}

// ⥸塼ν꡼󥢥å״ؿϿ
module_init(adc_init);
module_exit(adc_cleanup);
