#include <linux/config.h>
#include <linux/device.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/interrupt.h>    
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>

#include <asm/delay.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/ioctl.h>
#include <asm/arch/regs-adc.h>
#include <asm/arch/regs-gpio.h>
#include <asm/hardware/clock.h>


#define	SBC2440_ADC_SAMPLE_MAJOR 236         

MODULE_LICENSE("GPL");
MODULE_AUTHOR("zqing - FriendlyARM"); 
MODULE_DESCRIPTION("sbc2440_adc_sample");

static struct clk	*adc_clock;

static unsigned char *msg = "sbc2440_adc.o";

static unsigned int adc_ch = 0x00;

// ANZX
static int f_access = 0;

// ǂݏo|C^
static int read_p = 0;

static void __iomem *base_addr;

static int sbc2440_adc_sample_open(struct inode *inode, struct file *file)
{
	// 1vZXI[vłȂ悤ɂ
	if( f_access ) {
		printk( "%s: ADC is already used.", msg );
		return -EFAULT;
	}
	
	adc_ch = MINOR(inode->i_rdev);
	if( adc_ch > 3 ) {
		printk( "%s: ch > 3 is not supported.", msg );
		return -EFAULT;
	}

	read_p = 0;
	f_access++;
	return 0;
};
static int sbc2440_adc_sample_release(struct inode *inode, struct file *file)
{
	f_access--;
	return 0;
};

static short ReadAdc(int ch)
{
    short tmp;

	tmp = readl(base_addr+S3C2410_ADCCON);
	tmp &= ~(7 << 3);
	writel(tmp|(ch<<3), base_addr+S3C2410_ADCCON);

	printk(KERN_ERR "Start ADC %d\n", ch);
	writel(readl(base_addr+S3C2410_ADCCON) | S3C2410_ADCCON_ENABLE_START, base_addr+S3C2410_ADCCON);

    while(readl(base_addr+S3C2410_ADCCON) & S3C2410_ADCCON_ENABLE_START);	//check if Enable_start is low
    while(!(readl(base_addr+S3C2410_ADCCON) & S3C2410_ADCCON_ECFLG));	//check if EC(End of Conversion) flag is high

	tmp = readl(base_addr+S3C2410_ADCDAT0) & 0x3ff;
	printk( KERN_INFO "S3C2410_ADC %d=> %04x \n", ch, tmp );
  	return tmp;
}

static ssize_t sbc2440_adc_sample_read(struct file *file, char *buf, size_t size, loff_t *loff)
{
	ssize_t retval = 0;
	unsigned char *p;
	unsigned short ad_data;

	if( read_p == 0 ) {
		ad_data = ReadAdc(adc_ch);
	}
	
	if( size >= 2 ) {
		retval = put_user(ad_data, (unsigned short __user *)buf)?: sizeof(short);
		read_p = 2;
	}
	else if( size != 0 ) {
		if( read_p == 0 ) {
			retval = put_user(ad_data, (unsigned char __user *)buf)?: sizeof(char);
			read_p++;
		}
		else {
			p = (unsigned char *)&ad_data;
			p += read_p;
			retval = put_user(*p, (unsigned char __user *)buf)?: sizeof(char);
			read_p++;
		}
	}
	if( read_p > 1 ){
		read_p = 0;
	}
	return retval;
};
static ssize_t sbc2440_adc_sample_write(struct file *file, const char *buf, size_t size, loff_t *loff)
{
    return 0;
};
int sbc2440_adc_sample_ioctl(struct inode *inode,    struct file *file,    unsigned int ioctl_num,    unsigned long ioctl_param)
{
    return 0;
};

static struct file_operations sbc2440_adc_sample_fops = {
    .read             = sbc2440_adc_sample_read,
    .write             = sbc2440_adc_sample_write,
    .ioctl             = sbc2440_adc_sample_ioctl,
    .open             = sbc2440_adc_sample_open,
    .release         = sbc2440_adc_sample_release,
};

static int __init sbc2440_adc_sample_module_init(void)
{
	short tmp;
	adc_clock = clk_get(NULL, "adc");
	if (!adc_clock) {
		printk(KERN_ERR "failed to get adc clock source\n");
		return -ENOENT;
	}
	clk_use(adc_clock);
	clk_enable(adc_clock);

	disable_irq(IRQ_ADC);
	disable_irq(IRQ_TC);

	base_addr=ioremap(S3C2410_PA_ADC,0x20);
	if (base_addr == NULL) {
		printk(KERN_ERR "Failed to remap register block\n");
		return -ENOMEM;
	}
	else{
		printk(KERN_ERR "base_addr: %x\n", base_addr);
	}

	tmp = readl(base_addr+S3C2410_ADCCON);
	printk( KERN_INFO "S3C2410_ADCCON=> %04x \n", tmp );
	printk( KERN_INFO "S3C2410_ADCTSC=> %04x \n", readl(base_addr+S3C2410_ADCTSC) );
	printk( KERN_INFO "S3C2410_ADCDLY=> %04x \n", readl(base_addr+S3C2410_ADCDLY) );

	tmp &= ~(7 << 3);
	writel(tmp, base_addr+S3C2410_ADCCON);
	writel(0x58, base_addr+S3C2410_ADCTSC);
	writel(0xff, base_addr+S3C2410_ADCDLY);

    printk("init sbc2440_adc_sample module device\n");
    register_chrdev(SBC2440_ADC_SAMPLE_MAJOR, "sbc2440_adc_sample", &sbc2440_adc_sample_fops);
    return 0;
}

static void __exit sbc2440_adc_sample_module_cleanup(void)
{
	if (adc_clock) {
		clk_disable(adc_clock);
		clk_unuse(adc_clock);
		clk_put(adc_clock);
		adc_clock = NULL;
	}

	iounmap(base_addr);
    printk("cleanup sbc2440_adc_sample module device\n");
}

module_init(sbc2440_adc_sample_module_init);
module_exit(sbc2440_adc_sample_module_cleanup);
